"""
Utility functions for QCFractal.
"""

import socket


def find_port() -> int:
    sock = socket.socket()
    sock.bind(("", 0))
    host, port = sock.getsockname()
    return port


def is_port_open(ip: str, port: int) -> bool:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.settimeout(1)
    try:
        s.connect((ip, int(port)))
        s.shutdown(socket.SHUT_RDWR)
        return True
    except ConnectionRefusedError:
        return False
    finally:
        s.close()



try:
    import flask
    import dash
    import dash_core_components
    import dash_html_components
    import dash_bootstrap_components
    import dash_coreui_components
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        "Could not find all dashboard requirements, please `conda install qcfractal-dashboard -c conda-forge` to use the dashboard."
    )


from .app import app
from .layout import layout

app.layout = layout



import dash_bootstrap_components as dbc
import dash_coreui_components as coreui
import dash_html_components as html

from .pages import landing

pages = []
navbar_items = []


def add_page(page_data):
    navbar_items.append(page_data["navitem"])

    cond = coreui.AppRouteConditional(page_data["layout"], id=page_data["id"], route=page_data["route"])
    pages.append(cond)


# Add rows and pages
add_page(landing.data)

# Group1
navbar_items.append({"name": "Components", "title": True})

layout = html.Div(
    [
        coreui.AppHeader(
            [
                coreui.AppNavbarBrand(
                    full={"src": "/assets/images/logo.png", "width": 140, "height": 25, "alt": "QCArchive Logo"}
                ),
                coreui.AppSidebarToggler(id="AppSidebartogglermd", className="d-md-down-none", display="lg"),
            ],
            fixed=True,
        ),
        html.Div(
            [
                coreui.AppSidebar(
                    [
                        coreui.AppSidebarHeader(),
                        coreui.AppSidebarForm(),
                        coreui.AppSidebarNav(id="current-url", navConfig={"items": navbar_items}),
                        coreui.AppSidebarFooter(),
                    ],
                    fixed=True,
                    display="lg",
                ),
                html.Main(
                    [
                        coreui.AppBreadcrumb(appRoutes=[{"path": "/", "name": "Dashboard"}]),
                        dbc.Container(pages, id="page-content", fluid=True),
                    ],
                    className="main",
                ),
            ],
            className="app-body",
        ),
        coreui.AppFooter(),
    ],
    className="app",
)



import pandas as pd
import plotly.graph_objs as go
from plotly.colors import DEFAULT_PLOTLY_COLORS

import dash_bootstrap_components as dbc

from .connection import get_socket

_default_margin = {"t": 5, "b": 5, "r": 5, "l": 5}


def list_managers(status=None, modified_after=None):
    socket = get_socket()

    managers = socket.get_managers(status=status, modified_after=modified_after)
    cols = [
        "cluster",
        "username",
        "tag",
        "completed",
        "submitted",
        "failures",
        # "returned",
        "created_on",
        "modified_on",
        "programs",
        "procedures",
        # "name"
    ]

    df = pd.DataFrame(managers["data"])
    df = df[cols]
    df["programs"] = df["programs"].apply(lambda x: ", ".join(sorted(x)))
    df["procedures"] = df["procedures"].apply(lambda x: ", ".join(sorted(x)))
    df.columns = [x.title() for x in cols]
    df.sort_values("Completed", inplace=True, ascending=False)
    return dbc.Table.from_dataframe(df, style={"width": "100%"})


def manager_graph(status=None, modified_after=None):
    socket = get_socket()

    managers = socket.get_managers(status=status, modified_after=modified_after)
    df = pd.DataFrame(managers["data"])

    bars = []
    if df.shape[0] > 0:
        data = df.groupby("cluster")[["completed", "submitted", "failures"]].sum()
        data["error"] = data["failures"]
        data["running"] = data["submitted"] - data["completed"]

        bar_iter = [
            ("completed", DEFAULT_PLOTLY_COLORS[0]),
            ("running", DEFAULT_PLOTLY_COLORS[2]),
            ("error", DEFAULT_PLOTLY_COLORS[3]),
        ]

        data.sort_values("completed", inplace=True, ascending=False)
        for status, color in bar_iter:
            bars.append(go.Bar(name=status.title(), x=data.index, y=data[status], marker_color=color))

    return go.Figure(
        data=bars,
        layout={
            "yaxis_type": "log",
            "barmode": "group",
            "yaxis": {"title": "Tasks"},
            "xaxis": {"title": "Cluster"},
            "margin": _default_margin,
        },
    )


def task_graph():

    socket = get_socket()

    cnts = socket.custom_query("task", "counts")
    df = pd.DataFrame(cnts["data"])

    df.loc[df["tag"].isna(), "tag"] = "None"
    order = df.groupby("tag")["count"].sum().sort_values(ascending=False).index

    bar_iter = [
        ("waiting", DEFAULT_PLOTLY_COLORS[0]),
        ("running", DEFAULT_PLOTLY_COLORS[2]),
        ("error", DEFAULT_PLOTLY_COLORS[3]),
    ]

    bars = []
    for status, color in bar_iter:
        bar_data = []
        for tag in order:
            matches = df[(df["status"] == status) & (df["tag"] == tag)]
            bar_data.append(matches["count"].sum())

        bars.append(go.Bar(name=status.title(), x=order, y=bar_data, marker_color=color))

    fig = go.Figure(
        data=bars,
        layout={
            "barmode": "group",
            "yaxis_type": "log",
            "yaxis": {"title": "Tasks"},
            "xaxis": {"title": "Tag"},
            "margin": _default_margin,
        },
    )

    return fig



import pandas as pd

import dash_bootstrap_components as dbc
import dash_core_components as dcc
import dash_coreui_components as coreui
from dash.dependencies import Input, Output
from flask import current_app

from ..app import app
from ..connection import get_socket
from ..dash_models import list_managers, manager_graph, task_graph

## Layout

top_row = dbc.Row(
    [
        dbc.Col(coreui.AppCard([dbc.CardHeader("Server Information"), dbc.CardBody(id="fractal-server-information")])),
        dbc.Col(
            coreui.AppCard([dbc.CardHeader("Database Information"), dbc.CardBody(id="fractal-database-information")]),
            # className="col-sm-6 col-md-3",
        ),
        dbc.Col(coreui.AppCard([dbc.CardHeader("Queue Information"), dbc.CardBody(id="fractal-queue-information")])),
    ],
    className="w-100",
)


@app.callback(
    [
        Output("fractal-server-information", "children"),
        Output("fractal-database-information", "children"),
        Output("fractal-queue-information", "children"),
    ],
    [Input("manager-overview-groupby", "value")],
)
def update_server_information(status):
    config = current_app.config["FRACTAL_CONFIG"]
    server = dcc.Markdown(
        f"""
**Name:** {config.fractal.name}

**Query Limit:** {config.fractal.query_limit}
            """
    )
    database = dcc.Markdown(
        f"""
**Name:** {config.database.database_name}

**Port:** {config.database.port}

**Host:** {config.database.host}
        """
    )

    queue = dcc.Graph(figure=task_graph())
    return server, database, queue


def managers_table(status):
    socket = get_socket()

    managers = socket.get_managers(status=status)
    df = pd.DataFrame(managers["data"])

    df["uuid"] = [x[:6] for x in df["uuid"]]
    cols = ["cluster", "username", "tag", "submitted", "completed", "failures"]

    table = dash_table.DataTable(
        id="table",
        columns=[{"name": i.title(), "id": i} for i in cols],
        data=df[cols].to_dict("records"),
        filtering=True,
        sorting=True,
    )
    return table


groupby_items = dcc.Checklist(
    id="manager-overview-groupby",
    options=[{"label": "ACTIVE", "value": "ACTIVE"}, {"label": "INACTIVE", "value": "INACTIVE"}],
    value=["ACTIVE"],
    labelStyle={"display": "inline-block"},
)

groupby_items2 = dcc.Checklist(
    id="manager-overview-groupby2",
    options=[{"label": "ACTIVE", "value": "ACTIVE"}, {"label": "INACTIVE", "value": "INACTIVE"}],
    value=["ACTIVE"],
    labelStyle={"display": "inline-block"},
)

managers = dbc.Row(
    [
        dbc.Col(
            [
                coreui.AppCard(
                    [
                        dbc.CardHeader("Manager Information"),
                        dbc.CardBody(
                            [
                                dbc.Col([dbc.Row([groupby_items])]),
                                dcc.Graph(id="manager-overview", style={"width": "100%"}),
                            ]
                        ),
                    ]
                )
            ],
            className="w-100",
        )
    ],
    className="w-100",
)

manager_list = dbc.Row(
    [
        dbc.Col(
            [
                coreui.AppCard(
                    [
                        dbc.CardHeader("Manager List"),
                        dbc.CardBody([dbc.Col([dbc.Row([groupby_items2])]), dbc.Col(id="manager-list")]),
                    ]
                )
            ],
            className="w-100",
        )
    ],
    className="w-100",
)


@app.callback(Output("manager-overview", "figure"), [Input("manager-overview-groupby", "value")])
def update_overview_graph(status):
    return manager_graph(status=status)


@app.callback(Output("manager-list", "children"), [Input("manager-overview-groupby2", "value")])
def update_overview_list(status):
    return list_managers(status=status)


### Build page return
layout = dbc.Row([top_row, managers, manager_list])

navbar = {
    "name": "Dashboard",
    "url": "/",
    "icon": "cui-speedometer icons",
    "badge": {
        "variant": "info",
        #        'text': 'NEW'
    },
}
data = {"layout": layout, "navitem": navbar, "id": "dashcard", "route": "/"}



from flask import current_app, g

from ..storage_sockets import storage_socket_factory


def get_socket():
    if "socket" not in g:
        config = current_app.config["FRACTAL_CONFIG"]
        g.socket = storage_socket_factory(config.database_uri(safe=False), max_limit=50000)

    return g.socket
